/* gui++.cpp 
 *
 * author: marco corvi <marco.corvi@elsag.it>
 * date  : Sept 2002
 *
 * C++ source file for the GUI functions
 */
#include "gui++.h"
#include <string.h> // strlen

void
GUI::setForeground( unsigned long col ) 
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, col);
  }
}

void 
GUI::drawButton( const char * msg, int x, int y, int w, int h )
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, 0xffffff );
    XSetBackground( theDisplay, theGC, 0xffffff );
    XFillRectangle( theDisplay, W, theGC, x, y, w, h );
    XSetForeground( theDisplay, theGC, 0x000000 );
    XDrawLine( theDisplay, W, theGC, x, y, x+w, y );
    XDrawLine( theDisplay, W, theGC, x+w, y, x+w, y+h );
    XDrawLine( theDisplay, W, theGC, x+w, y+h, x, y+h );
    XDrawLine( theDisplay, W, theGC, x, y+h, x, y );
    XDrawString(theDisplay, W, theGC, x+2, y+h-2, msg, strlen(msg) );
  }
}


void
GUI::drawPoint( int x, int y, unsigned long col )
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, col);
    XDrawPoint( theDisplay, W, theGC, x/m_zoom, y/m_zoom );
  }
}

void
GUI::drawCross( int x, int y, unsigned long col )
{
  if (do_graphics) {
    x /= m_zoom;
    y /= m_zoom;
    XSetForeground( theDisplay, theGC, col);
    XDrawLine( theDisplay, W, theGC, x-2, y, x+2, y );
    XDrawLine( theDisplay, W, theGC, x, y-2, x, y+2 );
  }
}

void
GUI::drawLine( int x1, int y1, int x2, int y2, unsigned long col )
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, col);
    XDrawLine( theDisplay, W, theGC,
               x1/m_zoom, y1/m_zoom, x2/m_zoom, y2/m_zoom );
  }
}

void
GUI::drawRectangle( int x, int y, int w, int h, unsigned long col )
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, col);
    XSetBackground( theDisplay, theGC, col);
    XFillRectangle( theDisplay, W, theGC,
                    x/m_zoom, y/m_zoom, w/m_zoom, h/m_zoom );
    XSetBackground( theDisplay, theGC, 0x000000);
  }
}

void
GUI::drawBox( int x, int y, int w, int h, unsigned long col )
{
  if (do_graphics) {
    XSetForeground( theDisplay, theGC, col);
    XSetBackground( theDisplay, theGC, col);
    XDrawRectangle( theDisplay, W, theGC,
                    x/m_zoom, y/m_zoom, w/m_zoom, h/m_zoom );
    XSetBackground( theDisplay, theGC, 0x000000);
  }
}



void 
GUI::displayString( const char * msg, int xoff, int yoff, 
   unsigned long col ) 
{
  if (do_graphics == 0) return;
  /*
   * FIXME: the region to erase need to be managed better
   */
  int len = strlen( msg );
  int w = xoff + 20 * len;
  int h = yoff;

  yoff -= 17;
  XSetForeground( theDisplay, theGC, 0x00000000);
  XFillRectangle( theDisplay, W, theGC, xoff, yoff, w-xoff, h-yoff );
  XSetForeground( theDisplay, theGC, col);
  XDrawString(theDisplay, W, theGC, xoff, h, msg, len);
}

void
GUI::displayImage(
	    unsigned char * img, 
		int d1, 
		int d2, 
		int xoff, 
		int yoff,
		int bpp      /* bytes_per_pixel : 4 = BGRA, 3 = RGB, 1 = Grey */ 
	)
{ 
  if (do_graphics == 0) return;

  int d1z = d1 / m_zoom;
  int d2z = d2 / m_zoom;

  // printf("Do Display Image %d %d \n", d1, d2);

  if (d1z != d1save || d2z != d2save) {
    if (imd != NULL) {
      // fprintf(stderr, "free imd %x \n", imd);
      free (imd);
    }
    /* N.B. size is 4 times d1*d2 */
    imd = (unsigned char *)malloc(4*d1z*d2z*sizeof(char));
    d1save = -1;
    d2save = -1;
  }

  if (imd == NULL) 
    // should throw
    return;

  memset( imd, 0, 4*d1z*d2z);
  d1save = d1z;
  d2save = d2z;

# ifdef DO_COLOR_LOOKUP
    for (int j=0; j<d2z; j++) 
      for (int i=0; i<d1z; i++) 
        *(imd+j*d1z+i) = (char) look_up[ *(img+(j*d1+i)*m_zoom) ];
# endif

  switch (bpp) {
  case 1:  // GREY
    for (int j=0; j<d2z; j++) {
      for (int i=0; i<d1z; i++) {
        unsigned char * chp = imd+4*(j*d1z+i);
        *(chp) = *(chp+1) = *(chp+2) = *(img+(j*d1+i)*m_zoom);
        *(chp+3) = 255;
    } }
	break;
  case 2:  // RGB 
    for (int j=0; j<d2z; j++) {
      for (int i=0; i<d1z; i++) {
        unsigned char * chp = imd+4*(j*d1z+i);
        int ji = 3*(j*d1+i)*m_zoom;
        *(chp  ) = *(img+ji);
        *(chp+1) = *(img+ji+1);
        *(chp+2) = *(img+ji+2);
        *(chp+3) = 255;
    } }
	break;
  case 4:  // BGRA
	// printf("display BRGA (%dx%d)\n", d1,d2);
    for (int j=0; j<d2z; j++) {
      for (int i=0; i<d1z; i++) {
        unsigned char * chp = imd+4*(j*d1z+i);
        int ji = 3*(j*d1+i)*m_zoom;
        *(chp  ) = *(img+ji+2);
        *(chp+1) = *(img+ji+0);
        *(chp+2) = *(img+ji+1);
        *(chp+3) = 255;
    } }
	break;
  default:
	// fprintf(stderr, "Wrong BPP %d \n", bpp);
    // should throw
	return;
  }
  // theData = (void *)imd;
  theData = (char *)imd;

  // if (theImage) {
  //   XInitImage(theImage);
  // } else {
    theImage = XCreateImage(theDisplay, theVisual, theDepth, ZPixmap, 0, 
                       theData, d1z, d2z, 32 /* 8 */, 0);
  // }

  #if 1
  if ( xoff+d1z > (int)windowWidth ) {
	windowWidth = xoff+d1z;
	if ( yoff+d2z > (int)windowHeight ) 
      windowHeight = yoff+d2z;
    XResizeWindow(theDisplay, W, windowWidth, windowHeight);
	// fprintf(stderr, "Resize Window (3) %d %d (offs %d %d - dims %d %d)\n",
	// 			   	windowWidth, windowHeight, xoff, yoff, d1z, d2z);
  } else if (yoff+d2z > (int)windowHeight ) {
	windowHeight = yoff+d2z;
	XResizeWindow(theDisplay, W, windowWidth, windowHeight);
	// fprintf(stderr, "Resize Window (4) %d %d (offs %d %d - dims %d %d)\n",
	// 			   	windowWidth, windowHeight, xoff, yoff, d1z, d2z);
  }
  #endif
  
  XPutImage(theDisplay, W, theGC, theImage, 0,0, xoff,yoff, d1z, d2z);
  // XMapWindow(theDisplay, W);
  XFlush(theDisplay);
  // fprintf(stderr, "Display Image done.\n");
}

void
GUI::resizeWindow(unsigned int d1, unsigned int d2) 
{
  #if 1
  if (do_graphics) 
    XResizeWindow(theDisplay, W, d1/m_zoom, d2/m_zoom);
  #endif
}

void
GUI::clearWindow( const char * fmt, ... ) 
{
  if (do_graphics) {
    char msg[256];
    if ( fmt == NULL ) {
      strcpy( msg, "GUI by marco corvi" );
    } else {
      va_list args;
      va_start( args, fmt);
      vsnprintf( msg, 255, fmt, args );
      va_end( args );
      msg[255] = 0;
    }
    XClearWindow( theDisplay, W );
    XStoreName( theDisplay, W, msg);
  }
}

/* ------------------------------------------------------------------- */
/*     constr. and destr.                                              */
/* ------------------------------------------------------------------- */
GUI::GUI( int ig, unsigned int w, unsigned int h, int zoom ) 
{
  if (w==0 || h==0) ig=0;
  m_zoom = zoom;
  do_graphics  = ig;
  windowWidth  = w/m_zoom;
  windowHeight = h/m_zoom + 30; // 30 pixel for the status bar
  imd=NULL;  /* display image */
  d1save=-1;
  d2save=-1;

  if (do_graphics == 0) return;

  theDisplay = XOpenDisplay(XDisplayName(NULL));
  int theScreen = XDefaultScreen(theDisplay);
  theGC      = XDefaultGC(theDisplay,theScreen);
  theVisual  = DefaultVisual(theDisplay,theScreen);
  theDepth   = XDefaultDepth(theDisplay,theScreen);   /* This becomes 8   */
  Colormap theCmap = XDefaultColormap(theDisplay,theScreen);
  Window rootW  = RootWindow(theDisplay,theScreen);

  XColor theBlack;
  XAllocNamedColor(theDisplay, theCmap, BGCOLOR, &theBlack, &theBlack);

  // printf("Drawable depth %d\n", theDepth);

  // Screen * pScreen = XDefaultScreenOfDisplay(theDisplay);
  // int theNumber    = (1L<<theDepth);               /* this becomes 256 */
  // int screenWidth  = XWidthOfScreen(pScreen);
  // int screenHeight = XHeightOfScreen(pScreen);
  // int theClass     = theVisual->class;
  // int * plane_mask;
  // XGCValues xgcv;

# ifdef DO_COLOR_LOOKUP
  Colormap cmap = XCreateColormap(theDisplay, rootW, theVisual, AllocNone);
  XColor    color[256];
  int lk_up = 0;
  int count = 0;

  for (int i=0; i<256; i++)
  { 
    color[i].flags = DoRed | DoGreen | DoBlue;
    color[i].red =   i* (65535 /255.0);
    color[i].green = i* (65535 /255.0);
    color[i].blue  = i* (65535 /255.0);
    count += XAllocColor(theDisplay, cmap, &(color[i]) );

    int j = 0;
    while ( (j<i) && (
            (color[i].red   != color[j].red) || 
            (color[i].green != color[j].green) ||
            (color[i].blue  != color[j].blue) ) )
      j++;
    if (j==i) 
    { 
      look_up[i]=lk_up;
      lk_up++;
    } else {
      look_up[i]=look_up[j];
    }
   
  }
# endif

  windowX = 0;
  windowY = 0;
  /*
  windowX = (screenWidth-windowWidth)>>1;
  windowY = (screenHeight-windowHeight)>>1;
  */

  XSynchronize(theDisplay, 1);

  XSetWindowAttributes xswa;

  xswa.background_pixmap = None;
  xswa.background_pixel  = theBlack.pixel;
  xswa.border_pixmap     = CopyFromParent;
  xswa.border_pixel      = theBlack.pixel;
  xswa.bit_gravity       = ForgetGravity;
  xswa.win_gravity       = NorthWestGravity;
  xswa.backing_store     = Always;
  xswa.backing_planes    = 0xffffffff;       /* default*/
  xswa.backing_pixel     = 0x00000000;       /* default*/
  xswa.save_under        = FALSE;
  xswa.event_mask        =
                           KeyPressMask |
                           KeyReleaseMask |
                           ButtonPressMask |
                           ButtonReleaseMask |
                           // ExposureMask |
                           // ResizeRedirectMask |
                         0 ;
  xswa.do_not_propagate_mask =
                           KeyPressMask |
                           KeyReleaseMask |
                           PointerMotionMask |
                           ButtonMotionMask |
                           // ResizeRedirectMask |
                         0 ;
  xswa.override_redirect = FALSE;

# ifdef DO_COLOR_LOOKUP
    xswa.colormap         = cmap;
# else
    xswa.colormap         = theCmap;
# endif

  unsigned long flag = 
         /* CWBackPixmap |*/
         CWBackPixel |
         /* CWBorderPixmap |*/
         /* CWBorderPixel |*/
         /* CWBitGravity |*/
         /* CWWinGravity |*/
         CWBackingStore |
         /* CWBackingPlanes |*/
         /* CWBackingPixel | */
         CWOverrideRedirect |
         CWSaveUnder |
         CWEventMask  |
         CWDontPropagate |
         CWColormap ;

  W = XCreateWindow(theDisplay, rootW, 
				windowX, windowY, windowWidth, windowHeight, 0, theDepth, 
                CopyFromParent, theVisual, flag, &xswa);

  // XSetWindowColormap(theDisplay, W, cmap);
  XSetWindowColormap(theDisplay, W, theCmap);
  // XResizeWindow(theDisplay, W, windowWidth,  windowHeight);
  XMapWindow(theDisplay, W);
}

GUI::~GUI() 
{ 
  if (do_graphics) {
    XDestroyWindow( theDisplay, W );
    // XFree( theData );
    // XDestroyImage( theImage );
    if (imd) free(imd);
  }
}

/* ------------------------------------------------------------------- */
/*     user intercation (mouse clicks and keys)                        */
/* ------------------------------------------------------------------- */
void
GUI::displayStatus( const char * fmt, ... )
{
  if (do_graphics == 0) return;
  if ( fmt == NULL )  return;

  char msg[256];
  va_list args;
  va_start( args, fmt);
  vsnprintf( msg, 255, fmt, args );
  va_end( args );
  msg[255] = 0;
  int len = strlen( msg );
  int h = windowHeight-17;
  XSetForeground( theDisplay, theGC, 0xcccccc);
  XFillRectangle( theDisplay, W, theGC, 
     0, windowHeight-30, windowWidth, windowHeight);
  XSetForeground( theDisplay, theGC, 0x000000);
  XDrawString(theDisplay, W, theGC, 20, h, msg, len);
}
  

unsigned int 
GUI::waitKeyPress( const char * msg ) {
  if (do_graphics == 0) return 0;
  displayStatus ( msg );

  while (1) {
    XNextEvent(theDisplay, &event);
    // fprintf(stderr, "Key type %d\n", event.type );
    if (event.type == KeyPress) 
      break;
  }
  // fprintf(stderr, "Key code %d\n", event.xkey.keycode);
  return event.xkey.keycode;
}

int
GUI::waitMsecAndMouseClick(int msec, int * x, int * y, const char * msg )
{
  if (do_graphics == 0) return 0;
  displayStatus ( msg );
  int s = msec / 1000;
  int m = msec % 1000;
  if (s>0) sleep(s);
  if (m>0) usleep( 1000*m );

  if (XCheckWindowEvent( theDisplay, W, ButtonRelease, &event)) {
    *x = event.xbutton.x * m_zoom;
    *y = event.xbutton.y * m_zoom;
    return event.xbutton.button;
  }
  return -1;
}


unsigned int 
GUI::waitMouseClick( const char * msg ) {
  if (do_graphics == 0) return 0;
  displayStatus ( msg );

  while (1) {
    XNextEvent(theDisplay, &event);
    if (event.type == 4) 
      break;
  }
  return event.xbutton.button;
}

unsigned int 
GUI::waitMouseClick(int * x, int * y, const char * msg) {
  if (do_graphics == 0) return 0;
  displayStatus ( msg );

  while (1) {
    XNextEvent(theDisplay, &event);
    if (event.type == 4) 
      break;
  }
  *x = event.xbutton.x * m_zoom;
  *y = event.xbutton.y * m_zoom;
  return event.xbutton.button;
}

unsigned int 
GUI::waitMsecsAndKey( int msec, const char * msg ) {
  /* 
   * should handle errors !!!!
   */
  if (do_graphics == 0) return 0;

  displayStatus ( msg );
  int s = msec / 1000;
  int m = msec % 1000;
  sleep(s);
  if (m>0)
    usleep( 1000*m );

  if (XCheckWindowEvent( theDisplay, W, KeyPress | KeyRelease, &event)) {
    return event.xkey.keycode;
  }
  return 0;
}
  
void
GUI::waitMsecs( int msec) {
  /* 
   * should handle errors !!!!
   */
  int s = msec / 1000;
  int m = msec % 1000;
  sleep(s);
  if (m>0)
    usleep( 1000*m );
}
 
#ifdef TEST

int main(int argc, char ** argv)
{
  GUI gui(1, 200, 100);
  gui.clearWindow( "TEST" );
  gui.drawPoint( 5,5 );
  gui.drawLine( 10,20, 10,50);
  gui.drawLine( 10,20, 30,20, 0xff0000);
  gui.displayString( "Press a key", 40,50 );

  int i = gui.waitKeyPress();
  printf("Pressed key %d\n", i);

  gui.displayString( "Click the mouse", 40,50 );
  int x,y;
  int j = gui.waitMouseClick( &x, &y );
  printf("Mouse click (button %d) at %d %d\n", j, x, y);

} 

#endif // TEST
